@@ -1,7 +1,10 @@
 import routes
-from twisted.internet.defer import Deferred
+from twisted.internet.defer import Deferred, inlineCallbacks
 from twisted.web.resource import Resource
 from twisted.web.server import NOT_DONE_YET
+from twisted.python.failure import Failure
+from twisted.python.log import logging
+
 
 class Dispatcher(Resource):
     '''
@@ -83,7 +86,7 @@ class Dispatcher(Resource):
 
         result = self.__mapper.match(environ=wsgi_environ)
 
-        handler = None
+        handler = self._render_404
 
         if result is not None:
             controller = result.get('controller', None)
@@ -95,24 +98,86 @@ class Dispatcher(Resource):
 
                 if action is not None:
                     del result['action']
-                    handler = getattr(controller, action, None)
+                    func = getattr(controller, action, None)
+                    if func:
+                        handler = lambda request: func(request, **result)
 
-        if handler:
-            response = handler(request, **result)
-
-            if isinstance(response, Deferred):
-                return NOT_DONE_YET
-            else:
-                return response
-
-        else:
-            return self._render_404(request)
+        self.__detect_and_execute_handler(request, handler)
+        return NOT_DONE_YET
 
+    # Subclasses can override with their own 404 rendering.
     def _render_404(self, request):
         request.setResponseCode(404)
         return '<html><head><title>404 Not Found</title></head>' \
                 '<body><h1>Not found</h1></body></html>'
 
+    # Subclasses can override with their own error rendering.
+    def _render_error(self, request, exception=None, failure=None):
+        return self.__render_default_error(request, exception, failure)
+
+    def __render_default_error(self, request, exception=None, failure=None):
+        if failure:
+            logging.error(failure.getTraceback())
+        else:
+            logging.exception(exception)
+        request.setResponseCode(500)
+        return '<html><head><title>500 Internal Server Error</title></head>' \
+                '<body><h1>Internal Server Error</h1></body></html>'
+
+    @inlineCallbacks
+    def __detect_and_execute_handler(self, request, handler, use_default_error_rendering=False):
+
+        # Detect the content and whether the request is complete based
+        # on what the handler returns.
+        try:
+            content = None
+            complete = False
+            response = handler(request)
+
+            if isinstance(response, Deferred):
+                content = yield response
+                complete = True
+
+            elif response is NOT_DONE_YET:
+                content = None
+                complete = False
+
+            else:
+                content = response
+                complete = True
+
+            # If this response is complete, but the request has not been
+            # finished yet, ensure finish is called.
+            if complete and not request.finished:
+                if content:
+                    request.write(content)
+                request.finish()
+
+        except Exception, e:
+
+            # When using inlineCallbacks, logger.exception() does not show the
+            # real traceback. Subclasses will need log failure.getTraceback()
+            # to show tracebacks. Use Failure._findFailure() to get the failure
+            # associated with this exception.
+            failure = None
+            try:
+                failure = Failure._findFailure()
+            except Exception:
+                failure = None
+
+            # Use default error rendering when a subclass override of
+            # _render_error itself raised an unhandled error.
+            if use_default_error_rendering:
+                self.__render_default_error(request, e, failure)
+
+            # Otherwise, render the error and finish the request. Prevent
+            # infinite recursion by ensuring this recursive invocation falls
+            # back to __render_default_error.
+            else:
+                handler = lambda request: self._render_error(request, e, failure)
+                yield self.__detect_and_execute_handler(request, handler,
+                        use_default_error_rendering=True)
+
 
 if __name__ == '__main__':
     import logging
